{-# OPTIONS --safe --without-K #-}

open import Data.Nat.Base using (ℕ)

module Regex.Properties (α : ℕ) where

open import Regex α
import Regex.Properties.Sub α as ⊑
import Regex.Properties.Super α as ⊒

open import Algebra.Definitions _≈_
open import Algebra.Structures _≈_
open import Data.Product.Base using (_,_)
open import Relation.Binary

------------------------------------------------------------------------
-- General properties

≈-refl : Reflexive _≈_
≈-refl = ⊑.refl , ⊒.refl

≈-sym : Symmetric _≈_
≈-sym (p , q) = q , p

≈-trans : Transitive _≈_
≈-trans (p , q) (r , s) = ⊑.trans p r , ⊒.trans q s

≈-isEquivalence : IsEquivalence _≈_
≈-isEquivalence = record
  { refl = ≈-refl
  ; sym = ≈-sym
  ; trans = ≈-trans
  }

setoid : Setoid _ _
setoid = record { isEquivalence = ≈-isEquivalence }

------------------------------------------------------------------------
-- Algebraic properties

-- Properties of _+_

+-cong : Congruent₂ _+_
+-cong (x⊑y , x⊒y) (u⊑v , u⊒v) = ⊑.+-cong x⊑y u⊑v , ⊒.+-cong x⊒y u⊒v

+-assoc : Associative _+_
+-assoc r₁ r₂ r₃ = ⊑.+-assoc r₁ r₂ r₃ , ⊒.+-assoc r₁ r₂ r₃

+-comm : Commutative _+_
+-comm r₁ r₂ = ⊑.+-comm r₁ r₂ , ⊒.+-comm r₁ r₂

+-idem : Idempotent _+_
+-idem r = ⊑.+-idem r , ⊒.+-idem r

+-identityˡ : LeftIdentity Λ _+_
+-identityˡ r = ⊑.+-identityˡ r , ⊒.+-identityˡ r

+-identityʳ : RightIdentity Λ _+_
+-identityʳ r = ⊑.+-identityʳ r , ⊒.+-identityʳ r

+-identity : Identity Λ _+_
+-identity = +-identityˡ , +-identityʳ

-- Properties of _⨟_

⨟-cong : Congruent₂ _⨟_
⨟-cong (x⊑y , x⊒y) (u⊑v , u⊒v) = ⊑.⨟-cong x⊑y u⊑v , ⊒.⨟-cong x⊒y u⊒v

⨟-distribˡ-+ : _⨟_ DistributesOverˡ _+_
⨟-distribˡ-+ r₁ r₂ r₃ = ⊑.⨟-distribˡ-+ r₁ r₂ r₃ , ⊒.⨟-distribˡ-+ r₁ r₂ r₃

⨟-distribʳ-+ : _⨟_ DistributesOverʳ _+_
⨟-distribʳ-+ r₁ r₂ r₃ = ⊑.⨟-distribʳ-+ r₁ r₂ r₃ , ⊒.⨟-distribʳ-+ r₁ r₂ r₃

⨟-distrib-+ : _⨟_ DistributesOver _+_
⨟-distrib-+ = ⨟-distribˡ-+ , ⨟-distribʳ-+

⨟-assoc : Associative _⨟_
⨟-assoc r₁ r₂ r₃ = ⊑.⨟-assoc r₁ r₂ r₃ , ⊒.⨟-assoc r₁ r₂ r₃

⨟-zeroˡ : LeftZero Λ _⨟_
⨟-zeroˡ r = ⊑.⨟-zeroˡ r , ⊒.⨟-zeroˡ r

⨟-zeroʳ : RightZero Λ _⨟_
⨟-zeroʳ r = ⊑.⨟-zeroʳ r , ⊒.⨟-zeroʳ r

⨟-zero : Zero Λ _⨟_
⨟-zero = ⨟-zeroˡ , ⨟-zeroʳ

⨟-identityˡ : LeftIdentity ε _⨟_
⨟-identityˡ r = ⊑.⨟-identityˡ r , ⊒.⨟-identityˡ r

⨟-identityʳ : RightIdentity ε _⨟_
⨟-identityʳ r = ⊑.⨟-identityʳ r , ⊒.⨟-identityʳ r

⨟-identity : Identity ε _⨟_
⨟-identity = ⨟-identityˡ , ⨟-identityʳ

------------------------------------------------------------------------
-- Algebraic structures

+-isMagma : IsMagma _+_
+-isMagma = record
  { isEquivalence = ≈-isEquivalence
  ; ∙-cong = +-cong
  }

+-isSemigroup : IsSemigroup _+_
+-isSemigroup = record
  { isMagma = +-isMagma
  ; assoc = +-assoc
  }

+-isMonoid : IsMonoid _+_ Λ
+-isMonoid = record
  { isSemigroup = +-isSemigroup
  ; identity = +-identity
  }

+-isCommutativeMonoid : IsCommutativeMonoid _+_ Λ
+-isCommutativeMonoid = record
  { isMonoid = +-isMonoid
  ; comm = +-comm
  }

+-⨟-isSemiringWithoutAnnihilatingZero : IsSemiringWithoutAnnihilatingZero _+_ _⨟_ Λ ε
+-⨟-isSemiringWithoutAnnihilatingZero = record
  { +-isCommutativeMonoid = +-isCommutativeMonoid
  ; *-cong = ⨟-cong
  ; *-assoc = ⨟-assoc
  ; *-identity = ⨟-identity
  ; distrib = ⨟-distrib-+
  }

+-⨟-isSemiring : IsSemiring _+_ _⨟_ Λ ε
+-⨟-isSemiring = record
  { isSemiringWithoutAnnihilatingZero = +-⨟-isSemiringWithoutAnnihilatingZero
  ; zero = ⨟-zero
  }

+-⨟-isIdempotentSemiring : IsIdempotentSemiring _+_ _⨟_ Λ ε
+-⨟-isIdempotentSemiring = record
  { isSemiring = +-⨟-isSemiring
  ; +-idem = +-idem
  }

